@file:OptIn(ExperimentalContracts::class)

package ru.landrafhomyak.utility.reference_counter

import kotlin.contracts.ExperimentalContracts
import kotlin.contracts.InvocationKind
import kotlin.contracts.contract
import kotlin.jvm.JvmName
import ru.landgrafhomyak.utility.highlevel_try_finally.safeAutoClose1
import ru.landgrafhomyak.utility.highlevel_try_finally.safeAutoClose2

@CloseableReferenceCounter_Debug.RequiresExplicitDebug
public class CloseableReferenceCounter_Debug {
	public fun interface Observer {
		public fun observeState(instance: CloseableReferenceCounter_Debug, actions: String)
	}

	@RequiresOptIn(level = RequiresOptIn.Level.WARNING)
	@Retention(AnnotationRetention.BINARY)
	public annotation class RequiresExplicitDebug

	private val _dbgName: String
	private val _errMessage: String
	private val _logger: Observer?
	private val _value: _AtomicLong
	private val _id: Long

	public constructor(dbgName: String, errMessage: String, logger: Observer? = null) {
		_Platform.jvm_assertNotNull(dbgName, "param: dbgName")
		_Platform.jvm_assertNotNull(errMessage, "param: errMessage")

		this._dbgName = dbgName
		this._errMessage = dbgName
		this._logger = logger
		this._value = _AtomicLong(0L)
		@Suppress("RemoveRedundantQualifierName")
		this._id = CloseableReferenceCounter_Debug._nextId.getAndUpdate(Long::inc)
	}


	@Suppress("NOTHING_TO_INLINE")
	private inline fun _throwErrors(valueToCheck: Long) {
		when {
			valueToCheck >= 0 || valueToCheck == _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE -> {}
			valueToCheck < _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE -> throw RuntimeException("Too many references")
			valueToCheck > _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE -> throw RuntimeException(".decref called more times than .incref")
		}
	}

	@RequiresExplicitDebug
	@JvmName("throwErrors")
	public fun throwErrors() {
		this._throwErrors(this._value.get())
	}

	@JvmName("throwClosed")
	public fun throwClosed() {
		throw IllegalStateException(this._errMessage)
	}

	@JvmName("incref")
	public fun incref() {
		this._value.update { o ->
			if (o < 0) {
				this._throwErrors(o)
				this.throwClosed()
				this._throwErrors(o + 1)
			}
			return@update o + 1
		}
		this._logger?.observeState(this, "incref")
	}

	@JvmName("tryIncref\$kt")
	public inline fun <R> tryIncref(block: () -> R): R {
		contract {
			callsInPlace(block, InvocationKind.EXACTLY_ONCE)
		}
		this.incref()
		return safeAutoClose2(onError = this::decref, action = block)
	}

	@JvmName("assertNotClosed")
	public fun assertNotClosed() {
		if (this._value.get() < 0) this.throwClosed()
	}

	@JvmName("decref")
	public fun decref() {
		this._value.update { o ->
			if (o < 0) {
				this._throwErrors(o)
				this.throwClosed()
				this._throwErrors(o - 1)
			}
			return@update o - 1
		}
		this._logger?.observeState(this, "decref")
	}

	@JvmName("tryDecref\$kt")
	public inline fun <R> tryDecref(block: () -> R): R {
		contract {
			callsInPlace(block, InvocationKind.EXACTLY_ONCE)
		}
		this.assertNotClosed()
		return safeAutoClose2(onSuccess = this::decref, action = block)
	}

	@get:JvmName("isClosed")
	public val isClosed: Boolean
		get() {
			val state = this._value.get()
			this._throwErrors(state)
			return state < 0;
		}

	@JvmName("close")
	public fun close(errExistRefs: String) {
		_Platform.jvm_assertNotNull(errExistRefs, "param: errExistRefs")
		val state = _CloseableReferenceCounter_LowLevel.compareAndExchange(this._value, 0, _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE)
		this._throwErrors(state)
		when {
			state > 0 -> throw IllegalStateException(errExistRefs)
			state < 0 -> this.throwClosed()
		}
		this._logger?.observeState(this, "closed")
	}

	@JvmName("withRef\$kt")
	public inline fun <R> withRef(protected: () -> R): R {
		this.incref()
		return safeAutoClose1(finally = this::decref, action = protected)
	}

	override fun toString(): String {
		val refcntCached = this._value.get()
		val stateRepr: String
		@Suppress("LiftReturnOrAssignment")
		when {
			refcntCached >= 0 -> stateRepr = refcntCached.toString()
			refcntCached == _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE -> stateRepr = "closed"
			refcntCached < _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE -> stateRepr = "overflow"
			refcntCached > _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE -> stateRepr = "underflow"
			else -> throw Error("Unreachable")
		}
		return "<ref counter \"${this._dbgName}@${this._id}\" [${stateRepr}]>"
	}

	public companion object {
		private val _nextId = _AtomicLong(0L)
	}

	public object ObserveToStdout : Observer {
		override fun observeState(instance: CloseableReferenceCounter_Debug, actions: String) {
			_Platform.jvm_assertNotNull(instance, "param: instance")
			_Platform.jvm_assertNotNull(actions, "param: actions")
			print("${instance} ${actions}\n")
		}
	}
}